/*
Problem 09 - State Machine

type GetNextCharProc = function( curstate: UInt32 ): UInt32;

procedure StateMachineInit( var state_machine: Handle; states: UInt32; chars: UInt32 );
procedure AddTransition( state_machine: Handle; state1, state2: UInt32; first_char, last_char: UInt32 );
procedure RunStateMachine( state_machine: Handle; start_state: UInt32; GetNextChar: GetNextCharProc; var stop_state: UInt32 );

StateMachineInit creates a new, empty state machine ready to accept state transitions, and containing states states (numbers 1 to states) and chars characters (numbered 1 to chars).  All the state machine information must be stored in the real Mac memory manager handle - it will be disposed with DisposeHandle and that must release all memory, so dont store any extra information outside the handle.  Also, you must be able to deal with having multiple state machine instantiated simultaneously.

AddTransition adds a transition from state1 to state2 for characters between (inclusive) first_char and last_char.  When you get to state1 and get a character between first_char and last_char you should proceed to state2.  The new transition overrides and previous transitions for these characters from state1 (ie, if you get transition 1->2 for chars 1-10, and then 1->3 for chars 5-6, you now have transition 1->2 for chars 1-4 and 7-10 and transition 1->3 for chars 5-6).

RunStateMachine starts in start_state and calls GetNextChar repeatedly until either it returns 0 or until it returns a character for which there is no transition at the current state.  Each time it gets a character, the state is moved to the state for which there is a transition from that state with that character [blerg :-].  The final state is returned in stop_state.
*/

#include "Solution.h"

#define kMaxInt 0xffffffff;

#define TheState(state_machine) ((MyState *)(*(state_machine)))
#define TheStateTransitions(state_machine) ((MyStateTransition *)(*(state_machine) + sizeof(MyState)))

typedef struct MyStateTransition {
	UInt32 inputChar;
	UInt32 oldState;
	UInt32 newState;
} MyStateTransition;

typedef struct MyState {
	UInt32 numStates;
	UInt32 numChars;
	UInt32 numTransitions;
} MyState;

// Fill in your solution and then submit this folder

// Team Name: FILL IN YOUR TEAM NAME!

pascal void StateMachineInit( Handle  *state_machine, UInt32 states, UInt32 chars)
{
	MyState *stateInfo;
	MyStateTransition *stateArray;
	int i;

	*state_machine = NewHandleClear(sizeof(MyState) + states*chars*sizeof(MyStateTransition));
	if (0==*state_machine) DebugStr("\p mem alloc err");

	stateInfo = TheState(*state_machine);
	stateArray = TheStateTransitions(*state_machine);

	stateInfo->numStates = states;
	stateInfo->numChars = chars;
	stateInfo->numTransitions = 0;
	
	for (i=0; i<states*chars; i++) {
		stateArray->inputChar = kMaxInt;
		stateArray->oldState = kMaxInt;
		stateArray->newState = kMaxInt;
		stateArray++;
	}
}

pascal void AddTransition( Handle state_machine, UInt32 state1, UInt32 state2, UInt32 first_char, UInt32 last_char )
{
	MyState *stateInfo;
	MyStateTransition *nextState, *lastState, *p;
	UInt32 i;

	stateInfo = TheState(state_machine);
	lastState = nextState = TheStateTransitions(state_machine) + stateInfo->numTransitions;
	
	for (i=first_char; i<=last_char; i++) {
		Boolean foundIt = false;
		for (p = TheStateTransitions(state_machine); p<lastState; p++) {
			if ((p->oldState == state1) && (p->inputChar==i)) {
				p->newState = state2;
				foundIt = true;
				break;
			}
		}
		if (!foundIt) {
			nextState->inputChar = i;
			nextState->oldState = state1;
			nextState->newState = state2;
			nextState++;
		}
	}
	stateInfo->numTransitions = nextState - TheStateTransitions(state_machine);
}

pascal void RunStateMachine( Handle state_machine, UInt32 start_state, GetNextCharProc GetNextChar, UInt32 *stop_state)
{
	MyState *stateInfo;
	MyStateTransition *lastState, *p;
	UInt32 theInput,currentState;

	stateInfo = TheState(state_machine);
	lastState = TheStateTransitions(state_machine) + stateInfo->numTransitions;
	
	currentState = start_state;
	while (true) {
		Boolean foundIt;
		theInput = (*GetNextChar)(state_machine,currentState);
		for (p=TheStateTransitions(state_machine); p<lastState; p++) {
			foundIt = false;
			if ((p->oldState == currentState) && (p->inputChar == theInput)) {
				currentState = p->newState;
				foundIt=true;
				break;
			}
		}
		if (!foundIt) {
			*stop_state = currentState;
			return;
		}
	}
}
